"""
    Amqp Protocol
"""

import asyncio
import logging

from . import channel as amqp_channel
from . import constants as amqp_constants
from . import frame as amqp_frame
from . import exceptions
from . import version


logger = logging.getLogger(__name__)


class AmqpProtocol(asyncio.StreamReaderProtocol):
    """The AMQP protocol for asyncio.

    See http://docs.python.org/3.4/library/asyncio-protocol.html#protocols for more information
    on asyncio's protocol API.
    """

    CHANNEL_FACTORY = amqp_channel.Channel

    def __init__(self, *args, **kwargs):
        loop = kwargs.get('loop') or asyncio.get_event_loop()
        super().__init__(asyncio.StreamReader(loop=loop), self.client_connected, loop=loop)
        self._on_error_callback = kwargs.get('on_error')

        self.connecting = asyncio.Future()
        self.connection_closed = asyncio.Event()
        self.stop_now = asyncio.Future()
        self.is_open = False
        self.version_major = None
        self.version_minor = None
        self.server_properties = None
        self.server_mechanisms = None
        self.server_locales = None
        self.reader = None
        self.writer = None
        self.worker = None
        self.hearbeat = None
        self.channels = {}
        self.server_frame_max = None
        self.server_channel_max = None
        self.channels_max_id = 0
        self.channels_ids = list(range(1, amqp_constants.MAX_CHANNELS + 1))

    def client_connected(self, reader, writer):
        self.reader = reader
        self.writer = writer

    def connection_lost(self, exc):
        logger.warning("Connection lost exc=%r", exc)
        self.connection_closed.set()
        self.is_open = False
        self._close_channels(reply_code=None, reply_text='Connection lost')
        super().connection_lost(exc)

    @asyncio.coroutine
    def close(self, no_wait=False, timeout=None):
        """Close connection (and all channels)"""
        frame = amqp_frame.AmqpRequest(self.writer, amqp_constants.TYPE_METHOD, 0)
        frame.declare_method(
            amqp_constants.CLASS_CONNECTION, amqp_constants.CONNECTION_CLOSE)
        encoder = amqp_frame.AmqpEncoder(frame.payload)
        # we request a clean connection close
        encoder.write_short(0)
        encoder.write_shortstr('')
        encoder.write_short(0)
        encoder.write_short(0)
        frame.write_frame(encoder)
        if not no_wait:
            yield from self.wait_closed(timeout=timeout)

    @asyncio.coroutine
    def wait_closed(self, timeout=None):
        yield from asyncio.wait_for(self.connection_closed.wait(), timeout=timeout)

    @asyncio.coroutine
    def close_ok(self, frame):
        self.stop()
        logger.info("Recv close ok")

    @asyncio.coroutine
    def start_connection(self, host, port, login, password, virtualhost, ssl=False,
            login_method='AMQPLAIN', insist=False):
        """Initiate a connection at the protocol level
            We send `PROTOCOL_HEADER'
        """

        if ssl:
            # TODO
            logger.warning('ssl is not supported yet, falling back to non-secure connection')

        if login_method != 'AMQPLAIN':
            # TODO
            logger.warning('only AMQPLAIN login_method is supported, falling back to AMQPLAIN')

        self.writer.write(amqp_constants.PROTOCOL_HEADER)

        # Wait 'start' method from the server
        yield from self.dispatch_frame()

        client_properties = {
            'capabilities': {
                'consumer_cancel_notify': True,
                'connection.blocked': False,
            },
            'product_version': version.__version__,
            'product': version.__packagename__,
            'copyright': 'BSD',
        }
        auth = {
            'LOGIN': login,
            'PASSWORD': password,
        }

        # waiting reply start with credentions and co
        yield from self.start_ok(client_properties, 'AMQPLAIN', auth, self.server_locales[0])

        # wait for a "tune" reponse
        yield from self.dispatch_frame()

        tune_ok = {
            'channel_max': self.server_channel_max,
            'frame_max': self.server_frame_max,
            'heartbeat': self.hearbeat,
        }
        # "tune" the connexion with max channel, max frame, heartbeat
        yield from self.tune_ok(**tune_ok)

        # open a virtualhost
        yield from self.open(virtualhost, capabilities='', insist=insist)

        # wait for open-ok
        yield from self.dispatch_frame()

        # for now, we read server's responses asynchronously
        self.worker = asyncio.async(self.run())

    def stop(self):
        self.is_open = False
        self.connection_closed.set()
        self.stop_now.set_result(None)

    @asyncio.coroutine
    def get_frame(self):
        """Read the frame, and only decode its header

        """
        frame = amqp_frame.AmqpResponse(self.reader)
        yield from frame.read_frame()
        return frame

    @asyncio.coroutine
    def dispatch_frame(self, frame=None):
        """Dispatch the received frame to the corresponding handler"""

        method_dispatch = {
            (amqp_constants.CLASS_CONNECTION, amqp_constants.CONNECTION_CLOSE): self.server_close,
            (amqp_constants.CLASS_CONNECTION, amqp_constants.CONNECTION_CLOSE_OK): self.close_ok,
            (amqp_constants.CLASS_CONNECTION, amqp_constants.CONNECTION_TUNE): self.tune,
            (amqp_constants.CLASS_CONNECTION, amqp_constants.CONNECTION_START): self.start,
            (amqp_constants.CLASS_CONNECTION, amqp_constants.CONNECTION_OPEN_OK): self.open_ok,
        }
        if not frame:
            frame = yield from self.get_frame()
            #print("frame.channel {} class_id {}".format(frame.channel, frame.class_id))

        if frame.frame_type == amqp_constants.TYPE_HEARTBEAT:
            yield from self.reply_to_hearbeat(frame)
            return

        if frame.channel is not 0:
            try:
                yield from self.channels[frame.channel].dispatch_frame(frame)
            except KeyError:
                logger.info("Unknown channel %s", frame.channel)
            return

        if (frame.class_id, frame.method_id) not in method_dispatch:
            logger.info("frame {} {} is not handled".format(frame.class_id, frame.method_id))
            return
        yield from method_dispatch[(frame.class_id, frame.method_id)](frame)

    def release_channel_id(self, channel_id):
        """Called from the channel instance, it relase a previously used
        channel_id
        """
        self.channels_ids.append(channel_id)

    def _close_channels(self, reply_code, reply_text):
        for channel in self.channels.values():
            channel.connection_closed(reply_code, reply_text)

    @asyncio.coroutine
    def run(self):
        while not self.stop_now.done():
            try:
                yield from self.dispatch_frame()
            except exceptions.AmqpClosedConnection:
                logger.info("Close connection")
                self.stop_now.set_result(None)
            except Exception:
                logger.exception('error on dispatch')

    @asyncio.coroutine
    def reply_to_hearbeat(self, frame):
        logger.debug("replyin' to heartbeat")
        frame = amqp_frame.AmqpRequest(self.writer, amqp_constants.TYPE_HEARTBEAT, 0)
        request = amqp_frame.AmqpEncoder()
        frame.write_frame(request)

    # Amqp specific methods
    @asyncio.coroutine
    def start(self, frame):
        """Method sent from the server to begin a new connection"""
        response = amqp_frame.AmqpDecoder(frame.payload)

        self.version_major = response.read_octet()
        self.version_minor = response.read_octet()
        self.server_properties = response.read_table()
        self.server_mechanisms = response.read_longstr().split(' ')
        self.server_locales = response.read_longstr().split(' ')

    @asyncio.coroutine
    def start_ok(self, client_properties, mechanism, auth, locale):
        frame = amqp_frame.AmqpRequest(self.writer, amqp_constants.TYPE_METHOD, 0)
        frame.declare_method(
            amqp_constants.CLASS_CONNECTION, amqp_constants.CONNECTION_START_OK)
        request = amqp_frame.AmqpEncoder()
        request.write_table(client_properties)
        request.write_shortstr(mechanism)
        request.write_table(auth)
        request.write_shortstr(locale.encode())
        frame.write_frame(request)

    @asyncio.coroutine
    def server_close(self, frame):
        """The server is closing the connection"""
        response = amqp_frame.AmqpDecoder(frame.payload)
        reply_code = response.read_short()
        reply_text = response.read_shortstr()
        class_id = response.read_short()
        method_id = response.read_short()
        self.stop()
        logger.warning("Server closed connection: %s, code=%s, class_id=%s, method_id=%s",
            reply_text, reply_code, class_id, method_id)
        self._close_channels(reply_code, reply_text)

        if self._on_error_callback:
            if asyncio.iscoroutinefunction(self._on_error_callback):
                yield from self._on_error_callback(exceptions.ChannelClosed(reply_code, reply_text))
            else:
                self._on_error_callback(exceptions.ChannelClosed(reply_code, reply_text))

    @asyncio.coroutine
    def tune(self, frame):
        decoder = amqp_frame.AmqpDecoder(frame.payload)
        self.server_channel_max = decoder.read_short()
        self.server_frame_max = decoder.read_long()
        self.hearbeat = decoder.read_short()

    @asyncio.coroutine
    def tune_ok(self, channel_max, frame_max, heartbeat):
        frame = amqp_frame.AmqpRequest(self.writer, amqp_constants.TYPE_METHOD, 0)
        frame.declare_method(
            amqp_constants.CLASS_CONNECTION, amqp_constants.CONNECTION_TUNE_OK)
        encoder = amqp_frame.AmqpEncoder()
        encoder.write_short(channel_max)
        encoder.write_long(frame_max)
        encoder.write_short(heartbeat or 0)

        frame.write_frame(encoder)

    @asyncio.coroutine
    def secure_ok(self, login_response):
        pass

    @asyncio.coroutine
    def open(self, virtual_host, capabilities='', insist=False):
        """Open connection to virtual host."""
        frame = amqp_frame.AmqpRequest(self.writer, amqp_constants.TYPE_METHOD, 0)
        frame.declare_method(
            amqp_constants.CLASS_CONNECTION, amqp_constants.CONNECTION_OPEN)
        encoder = amqp_frame.AmqpEncoder()
        encoder.write_shortstr(virtual_host)
        encoder.write_shortstr(capabilities)
        encoder.write_bool(insist)

        frame.write_frame(encoder)

    @asyncio.coroutine
    def open_ok(self, frame):
        self.is_open = True
        logger.info("Recv open ok")

    #
    ## aioamqp public methods
    #

    @asyncio.coroutine
    def channel(self, **kwargs):
        """Factory to create a new channel

        """
        if not len(self.channels_ids):
            raise exceptions.NoChannelAvailable()
        channel_id = self.channels_ids.pop(0)
        channel = self.CHANNEL_FACTORY(self, channel_id, **kwargs)
        self.channels[channel_id] = channel
        yield from channel.open()
        return channel



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-four-python.html

"""

import asyncio
import aioamqp

import random
import sys


@asyncio.coroutine
def callback(body, envelope, properties):
    print("consumer {} recved {} ({})".format(envelope.consumer_tag, body, envelope.delivery_tag))

@asyncio.coroutine
def receive_log():
    try:
        transport, protocol = yield from aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return

    channel = yield from protocol.channel()
    exchange_name = 'direct_logs'
    # TODO let rabbitmq choose the queue name
    queue_name = 'queue-%s' % random.randint(0, 10000)

    yield from channel.exchange(exchange_name, 'direct')

    yield from asyncio.wait_for(channel.queue(queue_name, durable=False, auto_delete=True), timeout=10)

    severities = sys.argv[1:]
    if not severities:
        print("Usage: %s [info] [warning] [error]" % (sys.argv[0],))
        sys.exit(1)

    for severity in severities:
        yield from asyncio.wait_for(channel.queue_bind(exchange_name='direct_logs',
                                    queue_name=queue_name,
                                    routing_key=severity), timeout=10)

    print(' [*] Waiting for logs. To exit press CTRL+C')

    yield from asyncio.wait_for(channel.basic_consume(queue_name, callback=callback), timeout=10)
    yield from asyncio.Event().wait()

loop = asyncio.get_event_loop()
loop.run_until_complete(receive_log())



from .constants import MESSAGE_PROPERTIES


class Properties:
    """Class for basic message properties"""
    __slots__ = tuple(MESSAGE_PROPERTIES)

    def __init__(
            self, content_type=None, content_encoding=None, headers=None, delivery_mode=None,
            priority=None, correlation_id=None, reply_to=None, expiration=None, message_id=None,
            timestamp=None, type=None, user_id=None, app_id=None, cluster_id=None):
        self.content_type = content_type
        self.content_encoding = content_encoding
        self.headers = headers
        self.delivery_mode = delivery_mode
        self.priority = priority
        self.correlation_id = correlation_id
        self.reply_to = reply_to
        self.expiration = expiration
        self.message_id = message_id
        self.timestamp = timestamp
        self.type = type
        self.user_id = user_id
        self.app_id = app_id
        self.cluster_id = cluster_id



import asyncio
from urllib.parse import urlparse

from .protocol import AmqpProtocol

from .version import __version__
from .version import __packagename__


@asyncio.coroutine
def connect(host='localhost', port=5672, login='guest', password='guest',
            virtualhost='/', ssl=False, login_method='AMQPLAIN', insist=False,
            protocol_factory=AmqpProtocol, **kwargs):
    """Convenient method to connect to an AMQP broker

        @host:          the host to connect to
        @port:          broker port
        @login:         login
        @password:      password
        @virtualhost:   AMQP virtualhost to use for this connection
        @login_method:  AMQP auth method
        @insist:        Insist on connecting to a server

        @kwargs:        Arguments to be given to the protocol_factory instance

        Returns:        a tuple (transport, protocol) of an AmqpProtocol instance
    """
    if kwargs:
        transport, protocol = yield from asyncio.get_event_loop().create_connection(
        lambda: protocol_factory(**kwargs), host, port)

    else:
        transport, protocol = yield from asyncio.get_event_loop().create_connection(
        protocol_factory, host, port)

    yield from protocol.start_connection(host, port, login, password, virtualhost, ssl=ssl,
        login_method=login_method, insist=insist)

    return (transport, protocol)


@asyncio.coroutine
def from_url(url, login_method='AMQPLAIN', insist=False, protocol_factory=AmqpProtocol):
    """ Connect to the AMQP using a single url parameter and return the client.
    """
    url = urlparse(url)

    if url.scheme not in ('amqp', 'amqps'):
        raise ValueError('Invalid protocol %s, valid protocols are amqp or amqps' % url.scheme)

    transport, protocol = yield from connect(
        host=url.hostname or 'localhost',
        port=url.port or 5672,
        login=url.username or 'guest',
        password=url.password or 'guest',
        virtualhost=(url.path[1:] if len(url.path) > 1 else '/'),
        ssl=(url.scheme == 'amqps'),
        login_method=login_method,
        insist=insist,
        protocol_factory=protocol_factory)
    return protocol



"""
    aioamqp exceptions
"""


class AioamqpException(Exception):
    pass

class ConfigurationError(AioamqpException):
    pass

class AmqpClosedConnection(AioamqpException):
    pass

class SynchronizationError(AioamqpException):
    pass

class EmptyQueue(AioamqpException):
    pass


class NoChannelAvailable(AioamqpException):
    """There is no room left for more channels"""


class ChannelClosed(AioamqpException):
    def __init__(self, code=0, message='Channel is closed'):
        super().__init__(code, message)
        self.code = code
        self.message = message


class DuplicateConsumerTag(AioamqpException):
    def __repr__(self):
        return ('The consumer tag specified already exists for this '
                'channel: %s' % self.args[0])


class ConsumerCancelled(AioamqpException):
    def __repr__(self):
        return ('The consumer %s has been cancelled' % self.args[0])


class PublishFailed(AioamqpException):
    def __init__(self, delivery_tag):
        super().__init__(delivery_tag)
        self.delivery_tag = delivery_tag

    def __repr__(self):
        return 'Publish failed because a nack was received for delivery_tag {}'.format(
            self.delivery_tag)



"""
    Helper class to decode AMQP responses

AMQP Frame implementations


0      1         3         7                      size+7      size+8
+------+---------+---------+    +-------------+     +-----------+
| type | channel |   size  |    |   payload   |     | frame-end |
+------+---------+---------+    +-------------+     +-----------+
 octets   short     long         'size' octets          octet

The frame-end octet MUST always be the hexadecimal value %xCE

type:

Type = 1, "METHOD": method frame.
Type = 2, "HEADER": content header frame.
Type = 3, "BODY": content body frame.
Type = 4, "HEARTBEAT": heartbeat frame.


Method Payload

0          2           4
+----------+-----------+-------------- - -
| class-id | method-id | arguments...
+----------+-----------+-------------- - -
    short     short       ...

Content Payload

0          2        4           12               14
+----------+--------+-----------+----------------+------------- - -
| class-id | weight | body size | property flags | property list...
+----------+--------+-----------+----------------+------------- - -
   short     short    long long       short        remainder...

"""

import asyncio
import io
import struct
import socket

import os

from . import exceptions
from . import constants as amqp_constants
from .properties import Properties
from itertools import count
from decimal import Decimal


DUMP_FRAMES = False


class AmqpEncoder:

    def __init__(self, writer=None):
        self.payload = io.BytesIO()

    def write_table(self, data_dict):

        self.write_long(0)                  # the table length (set later)
        if data_dict is not None and len(data_dict):
            start = self.payload.tell()
            for key, value in data_dict.items():
                self.write_shortstr(key)
                self.write_value(value)
            table_length = self.payload.tell() - start
            self.payload.seek(start - 4)    # move before the long
            self.write_long(table_length)   # and set the table length
            self.payload.seek(0, os.SEEK_END)  # return at the end

    def write_value(self, value):
        if isinstance(value, (bytes, str)):
            self.payload.write(b'S')
            self.write_longstr(value)
        elif isinstance(value, bool):
            self.payload.write(b't')
            self.write_bool(value)
        elif isinstance(value, dict):
            self.payload.write(b'F')
            self.write_table(value)
        elif isinstance(value, int):
            self.payload.write(b'I')
            self.write_long(value)
        else:
            raise Exception("type({}) unsupported".format(type(value)))

    def write_bits(self, *args):
        """Write consecutive bools to one byte"""
        assert len(args) <= 8, "write_bits can only write 8 bits into one octet, sadly"
        byte_value = 0

        for arg_index, bit in enumerate(args):
            if bit:
                byte_value |= (1 << arg_index)

        self.write_octet(byte_value)

    def write_bool(self, value):
        self.payload.write(struct.pack('?', value))

    def write_octet(self, octet):
        self.payload.write(struct.pack('!B', octet))

    def write_short(self, short):
        self.payload.write(struct.pack('!H', short))

    def write_long(self, integer):
        self.payload.write(struct.pack('!I', integer))

    def write_long_long(self, longlong):
        self.payload.write(struct.pack('!Q', longlong))

    def _write_string(self, string):
        if isinstance(string, str):
            self.payload.write(string.encode())
        elif isinstance(string, bytes):
            self.payload.write(string)

    def write_longstr(self, string):
        self.write_long(len(string))
        self._write_string(string)

    def write_shortstr(self, string):
        self.write_octet(len(string))
        self._write_string(string)

    def write_message_properties(self, properties):

        properties_flag_value = 0
        if properties is None:
            self.write_short(0)
            return

        diff = set(properties.keys()) - set(amqp_constants.MESSAGE_PROPERTIES)
        if diff:
            raise ValueError("%s are not properties, valid properties are %s" % (
                diff, amqp_constants.MESSAGE_PROPERTIES))

        start = self.payload.tell()                 # record the position
        self.write_short(properties_flag_value)     # set the flag later

        content_type = properties.get('content_type')
        if content_type:
            properties_flag_value |= amqp_constants.FLAG_CONTENT_TYPE
            self.write_shortstr(content_type)
        content_encoding = properties.get('content_encoding')
        if content_encoding:
            properties_flag_value |= amqp_constants.FLAG_CONTENT_ENCODING
            self.write_shortstr(content_encoding)
        headers = properties.get('headers')
        if headers is not None:
            properties_flag_value |= amqp_constants.FLAG_HEADERS
            self.write_table(headers)
        delivery_mode = properties.get('delivery_mode')
        if delivery_mode is not None:
            properties_flag_value |= amqp_constants.FLAG_DELIVERY_MODE
            self.write_octet(delivery_mode)
        priority = properties.get('priority')
        if priority is not None:
            properties_flag_value |= amqp_constants.FLAG_PRIORITY
            self.write_octet(priority)
        correlation_id = properties.get('correlation_id')
        if correlation_id:
            properties_flag_value |= amqp_constants.FLAG_CORRELATION_ID
            self.write_shortstr(correlation_id)
        reply_to = properties.get('reply_to')
        if reply_to:
            properties_flag_value |= amqp_constants.FLAG_REPLY_TO
            self.write_shortstr(reply_to)
        expiration = properties.get('expiration')
        if expiration:
            properties_flag_value |= amqp_constants.FLAG_EXPIRATION
            self.write_shortstr(expiration)
        message_id = properties.get('message_id')
        if message_id:
            properties_flag_value |= amqp_constants.FLAG_MESSAGE_ID
            self.write_shortstr(message_id)
        timestamp = properties.get('timestamp')
        if timestamp is not None:
            properties_flag_value |= amqp_constants.FLAG_TIMESTAMP
            self.write_long_long(timestamp)
        type_ = properties.get('type')
        if type_:
            properties_flag_value |= amqp_constants.FLAG_TYPE
            self.write_shortstr(type_)
        user_id = properties.get('user_id')
        if user_id:
            properties_flag_value |= amqp_constants.FLAG_USER_ID
            self.write_shortstr(user_id)
        app_id = properties.get('app_id')
        if app_id:
            properties_flag_value |= amqp_constants.FLAG_APP_ID
            self.write_shortstr(app_id)
        cluster_id = properties.get('cluster_id')
        if cluster_id:
            properties_flag_value |= amqp_constants.FLAG_CLUSTER_ID
            self.write_shortstr(cluster_id)

        self.payload.seek(start)                    # move before the flag
        self.write_short(properties_flag_value)     # set the flag
        self.payload.seek(0, os.SEEK_END)


class AmqpDecoder:
    def __init__(self, reader):
        self.reader = reader

    def read_bit(self):
        return bool(self.read_octet())

    def read_octet(self):
        data = self.reader.read(1)
        return ord(data)

    def read_signed_octet(self):
        data = self.reader.read(1)
        return struct.unpack('!b', data)[0]

    def read_short(self):
        data = self.reader.read(2)
        return struct.unpack('!H', data)[0]

    def read_signed_short(self):
        data = self.reader.read(2)
        return struct.unpack('!h', data)[0]

    def read_long(self):
        data = self.reader.read(4)
        return struct.unpack('!I', data)[0]

    def read_signed_long(self):
        data = self.reader.read(4)
        return struct.unpack('!i', data)[0]

    def read_long_long(self):
        data = self.reader.read(8)
        return struct.unpack('!Q', data)[0]

    def read_signed_long_long(self):
        data = self.reader.read(8)
        return struct.unpack('!q', data)[0]

    def read_float(self):
        # XXX: This used to read & unpack '!d', which is a double, not a shorter float
        data = self.reader.read(4)
        return struct.unpack('!f', data)[0]

    def read_double(self):
        data = self.reader.read(8)
        return struct.unpack('!d', data)[0]

    def read_decimal(self):
        decimals = self.read_octet()
        value = self.read_signed_long()
        return Decimal(value) * (Decimal(10) ** -decimals)

    def read_shortstr(self):
        data = self.reader.read(1)
        string_len = struct.unpack('!B', data)[0]
        data = self.reader.read(string_len)
        return data.decode()

    def read_longstr(self):
        string_len = self.read_long()
        data = self.reader.read(string_len)
        return data.decode()

    def read_timestamp(self):
        # TODO: decode into datetime?
        return self.read_long_long()

    def read_table(self):
        """Reads an AMQP table"""
        table_len = self.read_long()
        table_data = AmqpDecoder(io.BytesIO(self.reader.read(table_len)))
        table = {}
        while table_data.reader.tell() < table_len:
            var_name = table_data.read_shortstr()
            var_value = self.read_table_subitem(table_data)
            table[var_name] = var_value

        return table

    def read_table_subitem(self, table_data):
        """Read `table_data` bytes, guess the type of the value, and cast it.

            table_data:     a pair of b'<type><value>'
        """
        value_type = chr(table_data.read_octet())
        if value_type == 't':
            return table_data.read_bit()
        elif value_type == 'b':
            return table_data.read_octet()
        elif value_type == 'B':
            return table_data.read_signed_octet()
        elif value_type == 'U':
            return table_data.read_signed_short()
        elif value_type == 'u':
            return table_data.read_short()
        elif value_type == 'I':
            return table_data.read_signed_long()
        elif value_type == 'i':
            return table_data.read_long()
        elif value_type == 'L':
            return table_data.read_unsigned_long_long()
        elif value_type == 'l':
            return table_data.read_long_long()
        elif value_type == 'f':
            return table_data.read_float()
        elif value_type == 'd':
            return table_data.read_float()
        elif value_type == 'D':
            return table_data.read_decimal()
        elif value_type == 's':
            return table_data.read_shortstr()
        elif value_type == 'S':
            return table_data.read_longstr()
        elif value_type == 'A':
            return table_data.read_field_array()
        elif value_type == 'T':
            return table_data.read_timestamp()
        elif value_type == 'F':
            return table_data.read_table()
        elif value_type == 'V':
            return None
        else:
            raise ValueError('Unknown value_type {}'.format(value_type))

    def read_field_array(self):
        array_len = self.read_long()
        array_data = AmqpDecoder(io.BytesIO(self.reader.read(array_len)))
        field_array = []
        while array_data.reader.tell() < array_len:
            item = self.read_table_subitem(array_data)
            field_array.append(item)
        return field_array


class AmqpRequest:
    def __init__(self, writer, frame_type, channel):
        self.writer = writer
        self.frame_type = frame_type
        self.channel = channel
        self.class_id = None
        self.weight = None
        self.method_id = None
        self.payload = None
        self.next_body_size = None

    def declare_class(self, class_id, weight=0):
        self.class_id = class_id
        self.weight = 0

    def set_body_size(self, size):
        self.next_body_size = size

    def declare_method(self, class_id, method_id):
        self.class_id = class_id
        self.method_id = method_id

    def write_frame(self, encoder=None):
        payload = None
        if encoder is not None:
            payload = encoder.payload
        content_header = ''
        transmission = io.BytesIO()
        if self.frame_type == amqp_constants.TYPE_METHOD:
            content_header = struct.pack('!HH', self.class_id, self.method_id)
        elif self.frame_type == amqp_constants.TYPE_HEADER:
            content_header = struct.pack('!HHQ', self.class_id, self.weight, self.next_body_size)
        elif self.frame_type == amqp_constants.TYPE_BODY:
            # no specific headers
            pass
        elif self.frame_type == amqp_constants.TYPE_HEARTBEAT:
            # no specific headers
            pass
        else:
            raise Exception("frame_type {} not handled".format(self.frame_type))

        header = struct.pack('!BHI', self.frame_type, self.channel, payload.tell() + len(content_header))
        transmission.write(header)
        if content_header:
            transmission.write(content_header)
        if payload:
            transmission.write(payload.getvalue())
        transmission.write(amqp_constants.FRAME_END)
        return self.writer.write(transmission.getvalue())


class AmqpResponse:
    """Read a response from the AMQP server

    """
    def __init__(self, reader):
        self.reader = reader
        self.frame_type = None
        self.channel = 0  # default channel in AMQP
        self.payload_size = None
        self.frame_end = None
        self.frame_payload = None
        self.payload = None
        self.frame_class = None
        self.frame_method = None
        self.class_id = None
        self.method_id = None
        self.weight = None
        self.body_size = None
        self.property_flags = None
        self.properties = None
        self.arguments = {}
        self.frame_length = 0

        self.payload_decoder = None
        self.header_decoder = None

    @asyncio.coroutine
    def read_frame(self):
        """Decode the frame"""
        try:
            data = yield from self.reader.readexactly(7)
        except (asyncio.IncompleteReadError, socket.error) as ex:
            raise exceptions.AmqpClosedConnection() from ex

        frame_header = io.BytesIO(data)
        self.header_decoder = AmqpDecoder(frame_header)
        self.frame_type = self.header_decoder.read_octet()
        self.channel = self.header_decoder.read_short()
        self.frame_length = self.header_decoder.read_long()
        payload_data = yield from self.reader.readexactly(self.frame_length)

        if self.frame_type == amqp_constants.TYPE_METHOD:
            self.payload = io.BytesIO(payload_data)
            self.payload_decoder = AmqpDecoder(self.payload)
            self.class_id = self.payload_decoder.read_short()
            self.method_id = self.payload_decoder.read_short()

        elif self.frame_type == amqp_constants.TYPE_HEADER:
            self.payload = io.BytesIO(payload_data)
            self.payload_decoder = AmqpDecoder(self.payload)
            self.class_id = self.payload_decoder.read_short()
            self.weight = self.payload_decoder.read_short()
            self.body_size = self.payload_decoder.read_long_long()
            self.property_flags = 0
            for flagword_index in count(0):
                partial_flags = self.payload_decoder.read_short()
                self.property_flags |= partial_flags << (flagword_index * 16)
                if partial_flags & 1 == 0:
                    break
            decoded_properties = {}
            if self.property_flags & amqp_constants.FLAG_CONTENT_TYPE:
                decoded_properties['content_type'] = self.payload_decoder.read_shortstr()
            if self.property_flags & amqp_constants.FLAG_CONTENT_ENCODING:
                decoded_properties['content_encoding'] = self.payload_decoder.read_shortstr()
            if self.property_flags & amqp_constants.FLAG_HEADERS:
                decoded_properties['headers'] = self.payload_decoder.read_table()
            if self.property_flags & amqp_constants.FLAG_DELIVERY_MODE:
                decoded_properties['delivery_mode'] = self.payload_decoder.read_octet()
            if self.property_flags & amqp_constants.FLAG_PRIORITY:
                decoded_properties['priority'] = self.payload_decoder.read_octet()
            if self.property_flags & amqp_constants.FLAG_CORRELATION_ID:
                decoded_properties['correlation_id'] = self.payload_decoder.read_shortstr()
            if self.property_flags & amqp_constants.FLAG_REPLY_TO:
                decoded_properties['reply_to'] = self.payload_decoder.read_shortstr()
            if self.property_flags & amqp_constants.FLAG_EXPIRATION:
                decoded_properties['expiration'] = self.payload_decoder.read_shortstr()
            if self.property_flags & amqp_constants.FLAG_MESSAGE_ID:
                decoded_properties['message_id'] = self.payload_decoder.read_shortstr()
            if self.property_flags & amqp_constants.FLAG_TIMESTAMP:
                decoded_properties['timestamp'] = self.payload_decoder.read_long_long()
            if self.property_flags & amqp_constants.FLAG_TYPE:
                decoded_properties['type'] = self.payload_decoder.read_shortstr()
            if self.property_flags & amqp_constants.FLAG_USER_ID:
                decoded_properties['user_id'] = self.payload_decoder.read_shortstr()
            if self.property_flags & amqp_constants.FLAG_APP_ID:
                decoded_properties['app_id'] = self.payload_decoder.read_shortstr()
            if self.property_flags & amqp_constants.FLAG_CLUSTER_ID:
                decoded_properties['cluster_id'] = self.payload_decoder.read_shortstr()
            self.properties = Properties(**decoded_properties)

        elif self.frame_type == amqp_constants.TYPE_BODY:
            self.payload = payload_data

        elif self.frame_type == amqp_constants.TYPE_HEARTBEAT:
            pass

        else:
            raise ValueError("Message type {:x} not known".format(self.frame_type))
        self.frame_end = yield from self.reader.readexactly(1)
        assert self.frame_end == amqp_constants.FRAME_END

    def __str__(self):
        frame_data = {
            'type': self.frame_type,
            'channel': self.channel,
            'size': self.payload_size,
            'frame_end': self.frame_end,
            'payload': self.frame_payload,
        }
        output = """
0        1           3            7                        size+7        size+8
+--------+-----------+------------+    +---------------+     +--------------+
|{type!r:^8}|{channel!r:^11}|{size!r:^12}|    |{payload!r:^15}|     |{frame_end!r:^14}|
+--------+-----------+------------+    +---------------+     +--------------+
   type    channel       size                payload            frame-end
""".format(**frame_data)

        if self.frame_type == amqp_constants.TYPE_METHOD:
            method_data = {
                'class_id': self.class_id,
                'method_id': self.method_id,
            }
            type_output = """
0          2           4
+----------+-----------+-------------- - -
|{class_id:^10}|{method_id:^11}| arguments...
+----------+-----------+-------------- - -
  class-id   method-id       ...""".format(**method_data)

            output += os.linesep + type_output

        return output



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-five-python.html

"""

import asyncio
import aioamqp

import random
import sys


@asyncio.coroutine
def callback(body, envelope, properties):
    print("consumer {} recved {} ({})".format(envelope.consumer_tag, body, envelope.delivery_tag))


@asyncio.coroutine
def receive_log():
    try:
        transport, protocol = yield from aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return

    channel = yield from protocol.channel()
    exchange_name = 'topic_logs'
    # TODO let rabbitmq choose the queue name
    queue_name = 'queue-%s' % random.randint(0, 10000)

    yield from channel.exchange(exchange_name, 'topic')

    yield from asyncio.wait_for(channel.queue(queue_name, durable=False, auto_delete=True), timeout=10)

    binding_keys = sys.argv[1:]
    if not binding_keys:
        print("Usage: %s [binding_key]..." % (sys.argv[0],))
        sys.exit(1)

    for binding_key in binding_keys:
        yield from asyncio.wait_for(channel.queue_bind(exchange_name='topic_logs',
                                                       queue_name=queue_name,
                                                       routing_key=binding_key), timeout=10)

    print(' [*] Waiting for logs. To exit press CTRL+C')

    yield from asyncio.wait_for(channel.basic_consume(queue_name, callback=callback), timeout=10)


asyncio.get_event_loop().run_until_complete(receive_log())



"""
    Amqp channel specification
"""

import asyncio
import logging
import uuid
import io
from itertools import count

from . import constants as amqp_constants
from . import frame as amqp_frame
from . import exceptions
from .envelope import Envelope


logger = logging.getLogger(__name__)


class Channel:

    def __init__(self, protocol, channel_id, on_error=None):
        self.protocol = protocol
        self.channel_id = channel_id
        self.consumer_queues = {}
        self.consumer_callbacks = {}
        self._on_error_callback = on_error
        self.response_future = None
        self.close_event = asyncio.Event()
        self.cancelled_consumers = set()
        self.last_consumer_tag = None
        self.publisher_confirms = False
        self.delivery_tag_iter = None  # used for mapping delivered messages to publisher confirms

        self._futures = {}
        self._ctag_events = {}

    def _set_waiter(self, rpc_name):
        if rpc_name in self._futures:
            raise exceptions.SynchronizationError("Waiter already exists")

        fut = asyncio.Future()
        self._futures[rpc_name] = fut
        return fut

    def _get_waiter(self, rpc_name):
        fut = self._futures.pop(rpc_name, None)
        if not fut:
            raise exceptions.SynchronizationError("Call %s didn't set a waiter" % rpc_name)
        return fut

    @property
    def is_open(self):
        return not self.close_event.is_set()

    def _close_channel(self):
        self.protocol.release_channel_id(self.channel_id)
        self.close_event.set()

    def connection_closed(self, server_code=None, server_reason=None):
        for future in self._futures.values():
            kwargs = {}
            if server_code is not None:
                kwargs['code'] = server_code
            if server_reason is not None:
                kwargs['message'] = server_reason
            exc = exceptions.ChannelClosed(**kwargs)
            future.set_exception(exc)

        self._close_channel()

    @asyncio.coroutine
    def dispatch_frame(self, frame):
        methods = {
            (amqp_constants.CLASS_CHANNEL, amqp_constants.CHANNEL_OPEN_OK): self.open_ok,
            (amqp_constants.CLASS_CHANNEL, amqp_constants.CHANNEL_FLOW_OK): self.flow_ok,
            (amqp_constants.CLASS_CHANNEL, amqp_constants.CHANNEL_CLOSE_OK): self.close_ok,
            (amqp_constants.CLASS_CHANNEL, amqp_constants.CHANNEL_CLOSE): self.server_channel_close,

            (amqp_constants.CLASS_EXCHANGE, amqp_constants.EXCHANGE_DECLARE_OK): self.exchange_declare_ok,
            (amqp_constants.CLASS_EXCHANGE, amqp_constants.EXCHANGE_BIND_OK): self.exchange_bind_ok,
            (amqp_constants.CLASS_EXCHANGE, amqp_constants.EXCHANGE_UNBIND_OK): self.exchange_unbind_ok,
            (amqp_constants.CLASS_EXCHANGE, amqp_constants.EXCHANGE_DELETE_OK): self.exchange_delete_ok,

            (amqp_constants.CLASS_QUEUE, amqp_constants.QUEUE_DECLARE_OK): self.queue_declare_ok,
            (amqp_constants.CLASS_QUEUE, amqp_constants.QUEUE_DELETE_OK): self.queue_delete_ok,
            (amqp_constants.CLASS_QUEUE, amqp_constants.QUEUE_BIND_OK): self.queue_bind_ok,
            (amqp_constants.CLASS_QUEUE, amqp_constants.QUEUE_UNBIND_OK): self.queue_unbind_ok,
            (amqp_constants.CLASS_QUEUE, amqp_constants.QUEUE_PURGE_OK): self.queue_purge_ok,

            (amqp_constants.CLASS_BASIC, amqp_constants.BASIC_QOS_OK): self.basic_qos_ok,
            (amqp_constants.CLASS_BASIC, amqp_constants.BASIC_CONSUME_OK): self.basic_consume_ok,
            (amqp_constants.CLASS_BASIC, amqp_constants.BASIC_CANCEL_OK): self.basic_cancel_ok,
            (amqp_constants.CLASS_BASIC, amqp_constants.BASIC_GET_OK): self.basic_get_ok,
            (amqp_constants.CLASS_BASIC, amqp_constants.BASIC_GET_EMPTY): self.basic_get_empty,
            (amqp_constants.CLASS_BASIC, amqp_constants.BASIC_DELIVER): self.basic_deliver,
            (amqp_constants.CLASS_BASIC, amqp_constants.BASIC_CANCEL): self.server_basic_cancel,
            (amqp_constants.CLASS_BASIC, amqp_constants.BASIC_ACK): self.basic_server_ack,
            (amqp_constants.CLASS_BASIC, amqp_constants.BASIC_NACK): self.basic_server_nack,

            (amqp_constants.CLASS_CONFIRM, amqp_constants.CONFIRM_SELECT_OK): self.confirm_select_ok,
        }

        if (frame.class_id, frame.method_id) not in methods:
            raise NotImplementedError("Frame (%s, %s) is not implemented" % (frame.class_id, frame.method_id))
        yield from methods[(frame.class_id, frame.method_id)](frame)

    @asyncio.coroutine
    def _write_frame(self, frame, request, no_wait, timeout=None, no_check_open=False):
        if not self.is_open and not no_check_open:
            raise exceptions.ChannelClosed()
        frame.write_frame(request)

#
## Channel class implementation
#

    @asyncio.coroutine
    def open(self, timeout=None):
        """Open the channel on the server."""
        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_CHANNEL, amqp_constants.CHANNEL_OPEN)
        request = amqp_frame.AmqpEncoder()
        request.write_shortstr('')
        fut = self._set_waiter('open')
        try:
            yield from self._write_frame(frame, request, no_wait=False, timeout=timeout, no_check_open=True)
        except Exception:
            self._get_waiter('open')
            fut.cancel()
            raise
        yield from fut
        return fut.result()

    @asyncio.coroutine
    def open_ok(self, frame):
        self.close_event.clear()
        fut = self._get_waiter('open')
        fut.set_result(True)
        logger.debug("Channel openned")

    @asyncio.coroutine
    def close(self, reply_code=0, reply_text="Normal Shutdown", no_wait=False, timeout=None):
        """Close the channel."""
        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_CHANNEL, amqp_constants.CHANNEL_CLOSE)
        request = amqp_frame.AmqpEncoder()
        request.write_short(reply_code)
        request.write_shortstr(reply_text)
        request.write_short(0)
        request.write_short(0)
        frame.write_frame(request)
        if not no_wait:
            future = self._set_waiter('close')
            return (yield from future)

    @asyncio.coroutine
    def close_ok(self, frame):
        self._get_waiter('close').set_result(True)
        logger.info("Channel closed")
        self._close_channel()

    @asyncio.coroutine
    def server_channel_close(self, frame):
        results = {
            'reply_code': frame.payload_decoder.read_short(),
            'reply_text': frame.payload_decoder.read_shortstr(),
            'class_id': frame.payload_decoder.read_short(),
            'method_id': frame.payload_decoder.read_short(),
        }
        self.connection_closed(results['reply_code'], results['reply_text'])

    @asyncio.coroutine
    def flow(self, active, timeout=None):
        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_CHANNEL, amqp_constants.CHANNEL_FLOW)
        request = amqp_frame.AmqpEncoder()
        request.write_bits(active)
        fut = self._set_waiter('flow')
        try:
            yield from self._write_frame(frame, request, no_wait=False, timeout=timeout, no_check_open=True)
        except Exception:
            self._get_waiter('flow')
            fut.cancel()
            raise
        yield from fut
        return fut.result()

    @asyncio.coroutine
    def flow_ok(self, frame):
        decoder = amqp_frame.AmqpDecoder(frame.payload)
        active = bool(decoder.read_octet())
        self.close_event.clear()
        fut = self._get_waiter('flow')
        fut.set_result({'active': active})

        logger.debug("Flow ok")

#
## Exchange class implementation
#

    @asyncio.coroutine
    def exchange_declare(self, exchange_name, type_name, passive=False, durable=False,
                         auto_delete=False, no_wait=False, arguments=None, timeout=None):
        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_EXCHANGE, amqp_constants.EXCHANGE_DECLARE)
        request = amqp_frame.AmqpEncoder()
        # short reserved-1
        request.write_short(0)
        request.write_shortstr(exchange_name)
        request.write_shortstr(type_name)

        internal = False  # internal: deprecated
        request.write_bits(passive, durable, auto_delete, internal, no_wait)
        request.write_table(arguments)

        if not no_wait:
            future = self._set_waiter('exchange_declare')
            try:
                yield from self._write_frame(frame, request, no_wait, timeout=timeout)
            except Exception:
                self._get_waiter('exchange_declare')
                future.cancel()
                raise
            return (yield from future)

        yield from self._write_frame(frame, request, no_wait, timeout=timeout)

    @asyncio.coroutine
    def exchange_declare_ok(self, frame):
        future = self._get_waiter('exchange_declare')
        future.set_result(True)
        logger.debug("Exchange declared")
        return future

    @asyncio.coroutine
    def exchange_delete(self, exchange_name, if_unused=False, no_wait=False, timeout=None):
        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_EXCHANGE, amqp_constants.EXCHANGE_DELETE)
        request = amqp_frame.AmqpEncoder()
        # short reserved-1
        request.write_short(0)
        request.write_shortstr(exchange_name)
        request.write_bits(if_unused, no_wait)

        if not no_wait:
            future = self._set_waiter('exchange_delete')
            try:
                yield from self._write_frame(frame, request, no_wait, timeout=timeout)
            except Exception:
                self._get_waiter('exchange_delete')
                future.cancel()
                raise
            return (yield from future)

        yield from self._write_frame(frame, request, no_wait, timeout=timeout)

    @asyncio.coroutine
    def exchange_delete_ok(self, frame):
        future = self._get_waiter('exchange_delete')
        future.set_result(True)
        logger.debug("Exchange deleted")

    @asyncio.coroutine
    def exchange_bind(self, exchange_destination, exchange_source, routing_key,
                      no_wait=False, arguments=None, timeout=None):
        if arguments is None:
            arguments = {}
        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_EXCHANGE, amqp_constants.EXCHANGE_BIND)

        request = amqp_frame.AmqpEncoder()
        request.write_short(0)  # reserved
        request.write_shortstr(exchange_destination)
        request.write_shortstr(exchange_source)
        request.write_shortstr(routing_key)

        request.write_bits(no_wait)
        request.write_table(arguments)
        if not no_wait:
            future = self._set_waiter('exchange_bind')
            try:
                yield from self._write_frame(frame, request, no_wait, timeout=timeout)
            except Exception:
                self._get_waiter('exchange_bind')
                future.cancel()
                raise
            return (yield from future)

        yield from self._write_frame(frame, request, no_wait, timeout=timeout)

    @asyncio.coroutine
    def exchange_bind_ok(self, frame):
        future = self._get_waiter('exchange_bind')
        future.set_result(True)
        logger.debug("Exchange bound")

    @asyncio.coroutine
    def exchange_unbind(self, exchange_destination, exchange_source, routing_key, no_wait=False, arguments=None, timeout=None):
        if arguments is None:
            arguments = {}
        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.EXCHANGE_UNBIND, amqp_constants.EXCHANGE_UNBIND)

        request = amqp_frame.AmqpEncoder()
        request.write_short(0)  # reserved
        request.write_shortstr(exchange_destination)
        request.write_shortstr(exchange_source)
        request.write_shortstr(routing_key)

        request.write_bits(no_wait)
        request.write_table(arguments)
        if not no_wait:
            future = self._set_waiter('exchange_unbind')
            try:
                yield from self._write_frame(frame, request, no_wait, timeout=timeout)
            except Exception:
                self._get_waiter('exchange_unbind')
                future.cancel()
                raise
            return (yield from future)

        yield from self._write_frame(frame, request, no_wait, timeout=timeout)

    @asyncio.coroutine
    def exchange_unbind_ok(self, frame):
        future = self._get_waiter('exchange_unbind')
        future.set_result(True)
        logger.debug("Exchange bound")

#
## Queue class implementation
#

    @asyncio.coroutine
    def queue_declare(self, queue_name, passive=False, durable=False,
                      exclusive=False, auto_delete=False, no_wait=False, arguments=None, timeout=None):
        if arguments is None:
            arguments = {}

        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_QUEUE, amqp_constants.QUEUE_DECLARE)
        request = amqp_frame.AmqpEncoder()
        request.write_short(0)  # reserved
        request.write_shortstr(queue_name)
        request.write_bits(passive, durable, exclusive, auto_delete, no_wait)
        request.write_table(arguments)
        if not no_wait:
            future = self._set_waiter('queue_declare')
            try:
                yield from self._write_frame(frame, request, no_wait, timeout=timeout)
            except Exception:
                self._get_waiter('queue_declare')
                future.cancel()
                raise
            return (yield from future)

        yield from self._write_frame(frame, request, no_wait, timeout=timeout)

    @asyncio.coroutine
    def queue_declare_ok(self, frame):
        results = {
            'queue': frame.payload_decoder.read_shortstr(),
            'message_count': frame.payload_decoder.read_long(),
            'consumer_count': frame.payload_decoder.read_long(),
        }
        future = self._get_waiter('queue_declare')
        future.set_result(results)
        logger.debug("Queue declared")


    @asyncio.coroutine
    def queue_delete(self, queue_name, if_unused=False, if_empty=False, no_wait=False, timeout=None):
        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_QUEUE, amqp_constants.QUEUE_DELETE)

        request = amqp_frame.AmqpEncoder()
        request.write_short(0)  # reserved
        request.write_shortstr(queue_name)
        request.write_bits(if_unused, if_empty, no_wait)
        if not no_wait:
            future = self._set_waiter('queue_delete')
            try:
                yield from self._write_frame(frame, request, no_wait, timeout=timeout)
            except Exception:
                self._get_waiter('queue_delete')
                future.cancel()
                raise
            return (yield from future)

        yield from self._write_frame(frame, request, no_wait, timeout=timeout)

    @asyncio.coroutine
    def queue_delete_ok(self, frame):
        future = self._get_waiter('queue_delete')
        future.set_result(True)
        logger.debug("Queue deleted")

    @asyncio.coroutine
    def queue_bind(self, queue_name, exchange_name, routing_key, no_wait=False, arguments=None, timeout=None):
        """Bind a queue and a channel."""
        if arguments is None:
            arguments = {}
        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_QUEUE, amqp_constants.QUEUE_BIND)

        request = amqp_frame.AmqpEncoder()
        # short reserved-1
        request.write_short(0)
        request.write_shortstr(queue_name)
        request.write_shortstr(exchange_name)
        request.write_shortstr(routing_key)
        request.write_octet(int(no_wait))
        request.write_table(arguments)
        if not no_wait:
            future = self._set_waiter('queue_bind')
            try:
                yield from self._write_frame(frame, request, no_wait, timeout=timeout)
            except Exception:
                self._get_waiter('queue_bind')
                future.cancel()
                raise
            return (yield from future)

        yield from self._write_frame(frame, request, no_wait, timeout=timeout)

    @asyncio.coroutine
    def queue_bind_ok(self, frame):
        future = self._get_waiter('queue_bind')
        future.set_result(True)
        logger.debug("Queue bound")

    @asyncio.coroutine
    def queue_unbind(self, queue_name, exchange_name, routing_key, arguments=None, timeout=None):
        if arguments is None:
            arguments = {}
        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_QUEUE, amqp_constants.QUEUE_UNBIND)

        request = amqp_frame.AmqpEncoder()
        # short reserved-1
        request.write_short(0)
        request.write_shortstr(queue_name)
        request.write_shortstr(exchange_name)
        request.write_shortstr(routing_key)
        request.write_table(arguments)
        future = self._set_waiter('queue_unbind')
        try:
            yield from self._write_frame(frame, request, no_wait=False, timeout=timeout)
        except Exception:
            self._get_waiter('queue_unbind')
            future.cancel()
            raise
        return (yield from future)

    @asyncio.coroutine
    def queue_unbind_ok(self, frame):
        future = self._get_waiter('queue_unbind')
        future.set_result(True)
        logger.debug("Queue unbound")

    @asyncio.coroutine
    def queue_purge(self, queue_name, no_wait=False, timeout=None):
        frame = amqp_frame.AmqpRequest(self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_QUEUE, amqp_constants.QUEUE_PURGE)

        request = amqp_frame.AmqpEncoder()
        # short reserved-1
        request.write_short(0)
        request.write_shortstr(queue_name)
        request.write_octet(int(no_wait))
        if not no_wait:
            future = self._set_waiter('queue_purge')
            yield from self._write_frame(frame, request, no_wait, timeout=timeout)
            return (yield from future)

        yield from self._write_frame(frame, request, no_wait, timeout=timeout)

    @asyncio.coroutine
    def queue_purge_ok(self, frame):
        decoder = amqp_frame.AmqpDecoder(frame.payload)
        message_count = decoder.read_long()
        future = self._get_waiter('queue_purge')
        future.set_result({'message_count': message_count})

#
## Basic class implementation
#

    @asyncio.coroutine
    def basic_publish(self, payload, exchange_name, routing_key, properties=None, mandatory=False, immediate=False):
        assert len(payload) != 0, "Payload cannot be empty"
        if not self.is_open:
            raise exceptions.ChannelClosed()

        method_frame = amqp_frame.AmqpRequest(
            self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        method_frame.declare_method(
            amqp_constants.CLASS_BASIC, amqp_constants.BASIC_PUBLISH)

        method_request = amqp_frame.AmqpEncoder()
        method_request.write_short(0)
        method_request.write_shortstr(exchange_name)
        method_request.write_shortstr(routing_key)
        method_request.write_bits(mandatory, immediate)
        method_frame.write_frame(method_request)

        header_frame = amqp_frame.AmqpRequest(
            self.protocol.writer, amqp_constants.TYPE_HEADER, self.channel_id)
        header_frame.declare_class(amqp_constants.CLASS_BASIC)
        header_frame.set_body_size(len(payload))
        encoder = amqp_frame.AmqpEncoder()
        encoder.write_message_properties(properties)

        header_frame.write_frame(encoder)

        # split the payload

        frame_max = self.protocol.server_frame_max
        for chunk in (payload[0+i:frame_max+i] for i in range(0, len(payload), frame_max)):

            content_frame = amqp_frame.AmqpRequest(
                self.protocol.writer, amqp_constants.TYPE_BODY, self.channel_id)
            content_frame.declare_class(amqp_constants.CLASS_BASIC)
            encoder = amqp_frame.AmqpEncoder()
            if isinstance(chunk, str):
                encoder.payload.write(chunk.encode())
            else:
                encoder.payload.write(chunk)
            content_frame.write_frame(encoder)

        yield from self.protocol.writer.drain()

    @asyncio.coroutine
    def basic_qos(self, prefetch_size, prefetch_count, connection_global, timeout=None):
        frame = amqp_frame.AmqpRequest(
            self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_BASIC, amqp_constants.BASIC_QOS)
        request = amqp_frame.AmqpEncoder()
        request.write_long(prefetch_size)
        request.write_short(prefetch_count)
        request.write_bits(connection_global)

        future = self._set_waiter('basic_qos')
        yield from self._write_frame(frame, request, no_wait=False, timeout=timeout)
        return (yield from future)

    @asyncio.coroutine
    def basic_qos_ok(self, frame):
        future = self._get_waiter('basic_qos')
        future.set_result(True)
        logger.debug("Qos ok")


    @asyncio.coroutine
    def basic_client_nack(self, *args, **kwargs):
        pass

    @asyncio.coroutine
    def basic_server_nack(self, frame, delivery_tag=None):
        if delivery_tag is None:
            decoder = amqp_frame.AmqpDecoder(frame.payload)
            delivery_tag = decoder.read_long_long()
        fut = self._get_waiter('basic_server_ack_{}'.format(delivery_tag))
        logger.debug('Received nack for delivery tag {!r}'.format(delivery_tag))
        fut.set_exception(exceptions.PublishFailed(delivery_tag))

    @asyncio.coroutine
    def basic_recover_async(self):
        pass

    @asyncio.coroutine
    def basic_recover(self, *args, **kwargs):
        pass

    @asyncio.coroutine
    def basic_recover_ok(self, frame):
        pass

    @asyncio.coroutine
    def basic_consume(self, queue_name='', consumer_tag='', no_local=False, no_ack=False, exclusive=False,
                      no_wait=False, callback=None, arguments=None, on_cancel=None, timeout=None):
        # If a consumer tag was not passed, create one
        consumer_tag = consumer_tag or 'ctag%i.%s' % (self.channel_id, uuid.uuid4().hex)

        if arguments is None:
            arguments = {}

        if callback is None or not asyncio.iscoroutinefunction(callback):
            raise exceptions.ConfigurationError("basic_consume requires a coroutine callback")

        frame = amqp_frame.AmqpRequest(
            self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_BASIC, amqp_constants.BASIC_CONSUME)
        request = amqp_frame.AmqpEncoder()
        request.write_short(0)
        request.write_shortstr(queue_name)
        request.write_shortstr(consumer_tag)
        request.write_bits(no_local, no_ack, exclusive, no_wait)
        request.write_table(arguments)

        self.consumer_callbacks[consumer_tag] = callback
        self.last_consumer_tag = consumer_tag

        if not no_wait:
            future = self._set_waiter('basic_consume')
            try:
                yield from self._write_frame(frame, request, no_wait=no_wait, timeout=timeout)
            except Exception:
                self._get_waiter('basic_consume')
                future.cancel()
                raise
            yield from future
            self._ctag_events[consumer_tag].set()
            return future.result()

        yield from self._write_frame(frame, request, no_wait=no_wait, timeout=timeout)
        return {'consumer_tag': consumer_tag}

    @asyncio.coroutine
    def basic_consume_ok(self, frame):
        ctag = frame.payload_decoder.read_shortstr()
        results = {
            'consumer_tag': ctag,
        }
        future = self._get_waiter('basic_consume')
        future.set_result(results)
        self._ctag_events[ctag] = asyncio.Event()

    @asyncio.coroutine
    def basic_deliver(self, frame):
        response = amqp_frame.AmqpDecoder(frame.payload)
        consumer_tag = response.read_shortstr()
        delivery_tag = response.read_long_long()
        is_redeliver = response.read_bit()
        exchange_name = response.read_shortstr()
        routing_key = response.read_shortstr()
        content_header_frame = yield from self.protocol.get_frame()

        buffer = io.BytesIO()
        while(buffer.tell() < content_header_frame.body_size):
            content_body_frame = yield from self.protocol.get_frame()
            buffer.write(content_body_frame.payload)

        body = buffer.getvalue()
        envelope = Envelope(consumer_tag, delivery_tag, exchange_name, routing_key, is_redeliver)
        properties = content_header_frame.properties

        callback = self.consumer_callbacks[consumer_tag]

        event = self._ctag_events.get(consumer_tag)
        if event:
            yield from event.wait()
            del self._ctag_events[consumer_tag]
        yield from callback(body, envelope, properties)

    @asyncio.coroutine
    def server_basic_cancel(self, frame):
        """From the server, means the server won't send anymore messages to this consumer."""
        consumer_tag = frame.arguments['consumer_tag']
        self.cancelled_consumers.add(consumer_tag)
        logger.info("consume cancelled received")

    @asyncio.coroutine
    def basic_cancel(self, consumer_tag, no_wait=False, timeout=None):
        frame = amqp_frame.AmqpRequest(
            self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_BASIC, amqp_constants.BASIC_CANCEL)
        request = amqp_frame.AmqpEncoder()
        request.write_shortstr(consumer_tag)
        request.write_bits(no_wait)
        if not no_wait:
            future = self._set_waiter('basic_cancel')
            try:
                yield from self._write_frame(frame, request, no_wait=no_wait, timeout=timeout)
            except Exception:
                self._get_waiter('basic_cancel')
                future.cancel()
                raise
            return (yield from future)

        yield from self._write_frame(frame, request, no_wait=no_wait, timeout=timeout)

    @asyncio.coroutine
    def basic_cancel_ok(self, frame):
        results = {
            'consumer_tag': frame.payload_decoder.read_shortstr(),
        }
        future = self._get_waiter('basic_cancel')
        future.set_result(results)
        logger.debug("Cancel ok")

    @asyncio.coroutine
    def basic_get(self, queue_name='', no_ack=False, timeout=None):
        frame = amqp_frame.AmqpRequest(
            self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_BASIC, amqp_constants.BASIC_GET)
        request = amqp_frame.AmqpEncoder()
        request.write_short(0)
        request.write_shortstr(queue_name)
        request.write_bits(no_ack)
        future = self._set_waiter('basic_get')
        try:
            yield from self._write_frame(frame, request, no_wait=False, timeout=timeout)
        except Exception:
            self._get_waiter('basic_get')
            future.cancel()
            raise
        return (yield from future)

    @asyncio.coroutine
    def basic_get_ok(self, frame):
        data = {}
        decoder = amqp_frame.AmqpDecoder(frame.payload)
        data['delivery_tag'] = decoder.read_long_long()
        data['redelivered'] = bool(decoder.read_octet())
        data['exchange_name'] = decoder.read_shortstr()
        data['routing_key'] = decoder.read_shortstr()
        data['message_count'] = decoder.read_long()
        content_header_frame = yield from self.protocol.get_frame()

        buffer = io.BytesIO()
        while(buffer.tell() < content_header_frame.body_size):
            content_body_frame = yield from self.protocol.get_frame()
            buffer.write(content_body_frame.payload)

        data['message'] = buffer.getvalue()
        future = self._get_waiter('basic_get')
        future.set_result(data)

    @asyncio.coroutine
    def basic_get_empty(self, frame):
        future = self._get_waiter('basic_get')
        future.set_exception(exceptions.EmptyQueue)

    @asyncio.coroutine
    def basic_client_ack(self, delivery_tag, multiple=False, timeout=None):
        frame = amqp_frame.AmqpRequest(
            self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_BASIC, amqp_constants.BASIC_ACK)
        request = amqp_frame.AmqpEncoder()
        request.write_long_long(delivery_tag)
        request.write_bits(multiple)
        yield from self._write_frame(frame, request, no_wait=False, timeout=timeout)

    @asyncio.coroutine
    def basic_server_ack(self, frame):
        decoder = amqp_frame.AmqpDecoder(frame.payload)
        delivery_tag = decoder.read_long_long()
        fut = self._get_waiter('basic_server_ack_{}'.format(delivery_tag))
        logger.debug('Received ack for delivery tag {!r}'.format(delivery_tag))
        fut.set_result(True)

    @asyncio.coroutine
    def basic_reject(self, delivery_tag, requeue=False, timeout=None):
        frame = amqp_frame.AmqpRequest(
            self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(
            amqp_constants.CLASS_BASIC, amqp_constants.BASIC_REJECT)
        request = amqp_frame.AmqpEncoder()
        request.write_long_long(delivery_tag)
        request.write_bits(requeue)
        yield from self._write_frame(frame, request, no_wait=False, timeout=timeout)

#
## convenient aliases
#
    queue = queue_declare
    exchange = exchange_declare

    @asyncio.coroutine
    def publish(self, payload, exchange_name, routing_key, properties=None, mandatory=False, immediate=False):
        assert len(payload) != 0, "Payload cannot be empty"
        if not self.is_open:
            raise exceptions.ChannelClosed()

        if self.publisher_confirms:
            delivery_tag = next(self.delivery_tag_iter)
            fut = self._set_waiter('basic_server_ack_{}'.format(delivery_tag))

        method_frame = amqp_frame.AmqpRequest(
            self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        method_frame.declare_method(
            amqp_constants.CLASS_BASIC, amqp_constants.BASIC_PUBLISH)

        method_request = amqp_frame.AmqpEncoder()
        method_request.write_short(0)
        method_request.write_shortstr(exchange_name)
        method_request.write_shortstr(routing_key)
        method_request.write_bits(mandatory, immediate)
        method_frame.write_frame(method_request)

        header_frame = amqp_frame.AmqpRequest(
            self.protocol.writer, amqp_constants.TYPE_HEADER, self.channel_id)
        header_frame.declare_class(amqp_constants.CLASS_BASIC)
        header_frame.set_body_size(len(payload))
        encoder = amqp_frame.AmqpEncoder()
        encoder.write_message_properties(properties)

        header_frame.write_frame(encoder)

        # split the payload

        frame_max = self.protocol.server_frame_max
        for chunk in (payload[0+i:frame_max+i] for i in range(0, len(payload), frame_max)):

            content_frame = amqp_frame.AmqpRequest(
                self.protocol.writer, amqp_constants.TYPE_BODY, self.channel_id)
            content_frame.declare_class(amqp_constants.CLASS_BASIC)
            encoder = amqp_frame.AmqpEncoder()
            if isinstance(chunk, str):
                encoder.payload.write(chunk.encode())
            else:
                encoder.payload.write(chunk)
            content_frame.write_frame(encoder)

        yield from self.protocol.writer.drain()

        if self.publisher_confirms:
            yield from fut

    @asyncio.coroutine
    def confirm_select(self, *, no_wait=False, timeout=None):
        if self.publisher_confirms:
            raise ValueError('publisher confirms already enabled')
        frame = amqp_frame.AmqpRequest(
            self.protocol.writer, amqp_constants.TYPE_METHOD, self.channel_id)
        frame.declare_method(amqp_constants.CLASS_CONFIRM, amqp_constants.CONFIRM_SELECT)
        request = amqp_frame.AmqpEncoder()
        request.write_shortstr('')

        if not no_wait:
            future = self._set_waiter('confirm_select')
            try:
                yield from self._write_frame(frame, request, no_wait=no_wait, timeout=timeout)
            except Exception:
                self._get_waiter('confirm_select')
                future.cancel()
                raise
            return (yield from future)

        yield from self._write_frame(frame, request, no_wait=no_wait, timeout=timeout)

    @asyncio.coroutine
    def confirm_select_ok(self, frame):
        self.publisher_confirms = True
        self.delivery_tag_iter = count(1)
        fut = self._get_waiter('confirm_select')
        fut.set_result(True)
        logger.debug("Confirm selected")



#!/usr/bin/env python

import asyncio
import aioamqp

@asyncio.coroutine
def callback(body, envelope, properties):
    print(body)

@asyncio.coroutine
def receive():
    try:
        transport, protocol = yield from aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return

    channel = yield from protocol.channel()
    queue_name = 'py2.queue'

    yield from asyncio.wait_for(channel.queue(queue_name, durable=False, auto_delete=True), timeout=10)

    yield from asyncio.wait_for(channel.basic_consume(queue_name, callback=callback), timeout=10)


asyncio.get_event_loop().run_until_complete(receive())
asyncio.get_event_loop().run_forever()


"""
    Delivery of messages to consumers
"""


class Envelope:
    """Class for basic deliver message fields"""
    __slots__ = ('consumer_tag', 'delivery_tag', 'exchange_name', 'routing_key', 'is_redeliver')

    def __init__(self, consumer_tag, delivery_tag, exchange_name, routing_key, is_redeliver):
        self.consumer_tag = consumer_tag
        self.delivery_tag = delivery_tag
        self.exchange_name = exchange_name
        self.routing_key = routing_key
        self.is_redeliver = is_redeliver



#!/usr/bin/env python

import asyncio
import aioamqp

import sys


@asyncio.coroutine
def exchange_routing():
    try:
        transport, protocol = yield from aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return


    channel = yield from protocol.channel()
    queue_name = "task_queue"
    message = ' '.join(sys.argv[1:]) or "Hello World!"
    message_properties = {
        'delivery_mode': 2
    }

    yield from asyncio.wait_for(channel.queue(queue_name, durable=True), timeout=10)

    yield from channel.publish("Message ", '', queue_name, properties=message_properties)
    print(" [x] Sent %r" % (message,))

    yield from transport.close()


asyncio.get_event_loop().run_until_complete(exchange_routing())



#!/usr/bin/env python

"""
    Amqp client exemple

"""
import asyncio
import aioamqp


@asyncio.coroutine
def produce():
    try:
        transport, protocol = yield from aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return

    queue_name = 'py2.queue'
    channel = yield from protocol.channel()
    yield from asyncio.wait_for(channel.queue(queue_name, durable=False, auto_delete=True), timeout=10)

    while True:
        yield from channel.publish("py3.message", '', queue_name)
        yield from asyncio.sleep(10)


asyncio.get_event_loop().run_until_complete(produce())



import os
import re
import setuptools
import sys

py_version = sys.version_info[:2]

_file_content = open(os.path.join(os.path.dirname(__file__), 'aioamqp', 'version.py')).read()

rex = re.compile(r"""version__ = '(.*)'.*__packagename__ = '(.*)'""", re.MULTILINE | re.DOTALL)
VERSION, PACKAGE_NAME = rex.search(_file_content).groups()
description = 'AMQP implementation using asyncio'

setuptools.setup(
    name=PACKAGE_NAME,
    version=VERSION,
    author="Polyconseil dev' team",
    author_email='opensource+aioamqp@polyconseil.fr',
    url='https://github.com/polyconseil/aioamqp',
    description=description,
    long_description=open('README.rst').read(),
    download_url='https://pypi.python.org/pypi/aioamqp',
    packages=[
        'aioamqp',
    ],
    extras_require={
        ':python_version=="3.3"': ['asyncio'],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
    ],
    platforms='all',
    license='BSD'
)



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-three-python.html

"""

import asyncio
import aioamqp

import sys


@asyncio.coroutine
def exchange_routing():
    try:
        transport, protocol = yield from aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return

    channel = yield from protocol.channel()
    exchange_name = 'logs'
    message = ' '.join(sys.argv[1:]) or "info: Hello World!"

    yield from channel.exchange(exchange_name, 'fanout')

    yield from channel.publish(message, exchange_name=exchange_name, routing_key='')
    print(" [x] Sent %r" % (message,))

    yield from transport.close()


asyncio.get_event_loop().run_until_complete(exchange_routing())



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-three-python.html

"""

import asyncio
import aioamqp
import random



@asyncio.coroutine
def callback(body, envelope, properties):
    print(body)


@asyncio.coroutine
def receive_log():
    try:
        transport, protocol = yield from aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return


    channel = yield from protocol.channel()
    exchange_name = 'logs'
    # TODO let rabbitmq choose the queue name
    queue_name = 'queue-%s' % random.randint(0, 10000)

    yield from channel.exchange(exchange_name, 'fanout')

    yield from asyncio.wait_for(channel.queue(queue_name, durable=False, auto_delete=True), timeout=10)

    yield from asyncio.wait_for(channel.queue_bind(exchange_name=exchange_name, queue_name=queue_name, routing_key=''), timeout=10)

    print(' [*] Waiting for logs. To exit press CTRL+C')

    yield from asyncio.wait_for(channel.basic_consume(queue_name, callback=callback), timeout=10)


asyncio.get_event_loop().run_until_complete(receive_log())
